(* ========================================================================= *)
(*  Permuted lists, finite permutations and quick sort.                      *)
(*                                                                           *)
(*  Author: Marco Maggesi                                                    *)
(*          University of Florence, Italy                                    *)
(*          http://www.math.unifi.it/~maggesi/                               *)
(*                                                                           *)
(*          (c) Copyright, Marco Maggesi, 2005-2007                          *)
(* ========================================================================= *)

loadt "Permutation/morelist.ml";;
loadt "Permutation/permuted.ml";;
loadt "Permutation/permutation.ml";;
loadt "Permutation/qsort.ml";;
