(* ========================================================================= *)
(* Load geometric algebra theory plus complex/quaternions application        *)
(*                                                                           *)
(*        (c) Copyright, Capital Normal University, China, 2018.             *)
(*     Authors: Liming Li, Zhiping Shi, Yong Guan, Guohui Wang, Sha Ma.      *)
(* ========================================================================= *)

loadt "Geometric_Algebra/geometricalgebra.ml";;
loadt "Geometric_Algebra/quaternions.ml";;
