loadt "miz3/Samples/samples.ml";;
loadt "miz3/Samples/sample.ml";;
loadt "miz3/Samples/talk.ml";;
loadt "miz3/Samples/drinker.ml";;
loadt "miz3/Samples/irrat2.ml";;
loadt "miz3/Samples/lagrange.ml";;
loadt "miz3/Samples/lagrange1.ml";;
loadt "miz3/Samples/icms.ml";;
loadt "miz3/Samples/other_mizs.ml";;
loadt "miz3/Samples/robbins.ml";;
loadt "miz3/Samples/forster.ml";;
loadt "miz3/Samples/luxury.ml";;
loadt "miz3/Samples/tobias.ml";;
